/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: FIFOManager.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#include <iostream>
#include <cstdio>
#include <string>

#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>

#include <pthread.h>
#include <errno.h>
#include <stdexcept>
#include <algorithm>

#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>


#include "FIFOManager.h"

using namespace std;
using namespace OpenSOAP;


const std::string TAG_ERR = "*** E ***::";

//FIFO̊ǗsȂǗNX

const std::string FIFOManagerSIG = "FIFOManager::"; // for LOG

const char FIFOManager::FILE_PATH_DELIMIT = '/';

//------------------------------------------------------------
// RXgN^
//------------------------------------------------------------
FIFOManager::FIFOManager(const std::string& fifo_dir_path, const int port)
{
#ifdef DEBUG
  cerr << FIFOManagerSIG << "FIFOManager(" << fifo_dir_path << ")" << endl;
#endif

  //===========================================
  // mutex
  //===========================================
  int status = pthread_mutex_init(&fifoInfoColl_mutex_, NULL);
  if (0 != status) {
    //throw OpenSOAPException();
    throw std::exception();
  }
  //ϐ
  status = pthread_cond_init(&fifoInfoColl_cond_, NULL);
  if (0 != status) {
    //throw OpenSOAPException();
    throw std::exception();
  }
  

  //FIFOt@CfBNg
  std::string fifoDirPath(fifo_dir_path);

  //wfBNg̃`FbN
  // E񒷂Pȏł邱
  // E΃pXwł邱('/'Ŏn܂)
  // EŌオ'/'ŏIĂȂ,Ō'/'tĂ
  //   (t@CtpXŊǗ邽)
  if (fifoDirPath.empty() || FILE_PATH_DELIMIT != fifoDirPath.at(0)) {
    cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: invalid FIFO directory : ["
	 << fifoDirPath << "]" << endl;
    // "invalid FIFO directory"
    throw std::exception();
    // ONX̐݌vKv 
  }
  if (FILE_PATH_DELIMIT != fifoDirPath.at(fifoDirPath.length()-1)) {
    fifoDirPath += FILE_PATH_DELIMIT;
  }
  //typedef struct direct Dir;
  typedef struct stat Stat;

  DIR* dp = NULL;
  //  Dir* dir = NULL;
  struct dirent *dir = NULL;
  
  //wfBNgJ
  if (NULL == (dp = opendir(fifoDirPath.c_str()))) {
    cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: cannot open directory.[" 
	 << fifoDirPath << "]" << endl;
    // "open dir failed " + fifoDirPath
    throw std::exception();
    // ONX̐݌vKv 
  }

  Stat sbuf;
  string file_path;

  //vector<string> fifoVector;
  std::list<string> fifoColl;

  while (NULL != (dir = readdir(dp))) {
    if (0 == dir->d_ino) {
      cerr << "0 == dir->d_ino" << endl;
      continue;
    }
    //full path file name
    file_path = fifoDirPath + dir->d_name;
  
#ifdef DEBUG
    cout << "DIR:" << file_path << endl;
#endif

    //t@C擾
    if (0 > stat(file_path.c_str(), &sbuf)) { // stat is System call
      cerr << FIFOManagerSIG << " stat failed.[" 
	   << file_path << "]" << endl;
      //perror(file_path);
      cerr << "FIFOManager::FIFOManager: stat failed" << endl;
      continue;
    }
    //fBNg͑ΏۊO
    if (S_IFDIR != (sbuf.st_mode & S_IFMT)) { //{͂FIFOt@C̃`FbN
      //FIFOt@CK`FbN
      if (false == FIFOInfo::checkValidFileName(dir->d_name)) {
	//logging
	cerr << " invalid file name : [" << file_path << "] IGNORE" << endl;
	continue;
      }
      //fifoVector.push_back(file_path);
      fifoColl.push_back(file_path);
    }
  }

#ifdef DEBUG
  cerr << "FIFOManager:: read FIFO list:" << endl;
  cerr << "-----------------------------" << endl;
  //vector<string>::const_iterator pos;
  std::list<string>::const_iterator pos;
  for (pos = fifoColl.begin(); pos != fifoColl.end(); ++pos) {
    cerr << "\t" << *pos << endl;
  }
  cerr << "-----------------------------" << endl;
#endif

  closedir(dp);

  //FIFOInfoXg̐
  makeFIFOInfo(fifoColl);

  //===================================================================
  // ʐM FIFOManager <-> FIFOSelector
  //===================================================================
  
  //const int TCP_PORT = 6500;

  typedef struct sockaddr SockAddr;
  typedef struct sockaddr_in SockAddrIn;

  int sockfd = int();
  int newsockfd = int();
  SockAddrIn serv_addr;
  SockAddrIn client_addr;

  //FIFOSelectorƂ̒ʐM|[g
  if (0 > (sockfd = socket(PF_INET, SOCK_STREAM, 0))) {
    cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: open socket failed" << endl;
    throw std::exception();
    // ONX̐݌vKv 
  }
#ifdef DEBUG
    cerr << " socket open OK!" << endl;
#endif

  memset((char*)&serv_addr, 0, sizeof(serv_addr));
  serv_addr.sin_family = PF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  //serv_addr.sin_port = htons(TCP_PORT);
  serv_addr.sin_port = htons(port);

  if (0 > bind(sockfd, (SockAddr*)&serv_addr, sizeof(serv_addr))) {
    cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: bind failed" << endl;
    throw std::exception();
    // ONX̐݌vKv 
  }
#ifdef DEBUG
    cerr << " bind OK!" << endl;
#endif

  listen(sockfd, 5);

#ifdef DEBUG
  cerr << " listen OK!" << endl;
#endif
    
  socklen_t client_len = socklen_t();
  for (;;) {
    client_len = sizeof(client_addr);

#ifdef DEBUG
    cerr << " accept ready!" << endl;
#endif
    
    newsockfd = accept(sockfd, (SockAddr*)&client_addr, &client_len);
    if (0 > newsockfd) {
      cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: accept failed" << endl;
      throw std::exception();
      // ONX̐݌vKv 
    }

#ifdef DEBUG
    cerr << " accept OK!" << endl;
#endif

    //XbhɓnKv̂f[^
    // Eint newsockfd;
    // EFIFOManager* this;
    ThrData* thrData = new ThrData();
    thrData->that = this;
    thrData->sockfd = newsockfd;

    //ʐMXbh쐬
    pthread_t thr_id;
    int status = pthread_create(&thr_id,
				 NULL,
				 connectSelector,
				 (void*)thrData);
    if (0 != status) {
      cerr << TAG_ERR << FIFOManagerSIG << "FIFOManager: create thread failed" << endl;
      throw std::exception();
      // ONX̐݌vKv 
    }
  }
}


//=========================================
//FIFOSelectorƂ̒ʐMXbh\bh
//=========================================
void* FIFOManager::connectSelector(void* arg) {

  //ñ`FbN
  if (!arg) {
    cerr << TAG_ERR << FIFOManagerSIG << "connectSelector: invalid argument" << endl;
    pthread_exit(NULL);
  }

  //n̓WJ
  ThrData* thrData = (ThrData*)arg;
  FIFOManager* that = thrData->that;
  int sockfd = thrData->sockfd;
  delete thrData;

  char data[BUFSIZ];
  read(sockfd, data, sizeof(data));
  cerr << " connectSelector(" << sockfd << "):: read:[" << data << "]" << endl;


  //Mf[^̉
  const std::string delims(" \t");
  std::string recvData(data);
  std::string::size_type begIdx, endIdx;
  begIdx = recvData.find_first_not_of(delims);
  if (begIdx == std::string::npos) {
    cerr << "all blank" << endl;
    //return -1;
    pthread_exit(NULL);
  }
  endIdx = recvData.find_first_of(delims, begIdx);
  if (endIdx == std::string::npos) {
    endIdx = recvData.length();
  }
  std::string recvCommand = recvData.substr(begIdx, (endIdx-begIdx));
  cerr << "recvCommand=[" << recvCommand << "]" << endl;
  
  FIFOInfo* fifoInfo = NULL;

  if (OPEN_UPSIDE_FIFO == recvCommand) { 

    // timeout second
    begIdx = recvData.find_first_not_of(delims, endIdx);
    if (begIdx == std::string::npos) {
      cerr << "not recv timeout second" << endl;
      //return -1;
      pthread_exit(NULL);
    }
    endIdx = recvData.find_first_of(delims, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = recvData.length();
    }
    std::string timeoutSecondStr = recvData.substr(begIdx, (endIdx-begIdx));
    int timeoutSecond = atoi(timeoutSecondStr.c_str());
    cerr << "timeout=[" << timeoutSecondStr << "](" << timeoutSecond << ")" << endl;

    fifoInfo = that->openUpsideFIFO(timeoutSecond);
    if (!fifoInfo) {
      sprintf(data, "FAILURE 101");
      write(sockfd, data, sizeof(data));
    } else {
      sprintf(data, "SUCCESS %s %s %s", 
	      fifoInfo->getUpsideFIFODescriptor().getId().c_str(),
	      fifoInfo->getUpsideFIFO().c_str(),
	      fifoInfo->getDownsideFIFO().c_str());
      write(sockfd, data, sizeof(data));
    }
      
  } else if (OPEN_DOWNSIDE_FIFO == recvCommand) {

    // timeout second
    begIdx = recvData.find_first_not_of(delims, endIdx);
    if (begIdx == std::string::npos) {
      cerr << "not recv timeout second" << endl;
      //return -1;
      pthread_exit(NULL);
    }
    endIdx = recvData.find_first_of(delims, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = recvData.length();
    }
    std::string timeoutSecondStr = recvData.substr(begIdx, (endIdx-begIdx));
    int timeoutSecond = atoi(timeoutSecondStr.c_str());
    cerr << "timeout=[" << timeoutSecondStr << "](" << timeoutSecond << ")" << endl;

    fifoInfo = that->openDownsideFIFO(timeoutSecond);
    if (!fifoInfo) {
      sprintf(data, "FAILURE 102");
      write(sockfd, data, sizeof(data));
    } else {
      sprintf(data, "SUCCESS %s %s %s", 
	      fifoInfo->getDownsideFIFODescriptor().getId().c_str(),
	      fifoInfo->getUpsideFIFO().c_str(),
	      fifoInfo->getDownsideFIFO().c_str());
      write(sockfd, data, sizeof(data));
    }

  } else if (CLOSE_UPSIDE_FIFO == recvCommand) {
    // ID
    begIdx = recvData.find_first_not_of(delims, endIdx);
    if (begIdx == std::string::npos) {
      cerr << "not recv id" << endl;
      //return -1;
      pthread_exit(NULL);
    }
    endIdx = recvData.find_first_of(delims, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = recvData.length();
    }
    std::string id = recvData.substr(begIdx, (endIdx-begIdx));
    cerr << "id=[" << id << "]" << endl;

    int result = that->closeUpsideFIFO(id);
    if (0 != result) {
      sprintf(data, "FAILURE 103");
      write(sockfd, data, sizeof(data));
    } else {
      sprintf(data, "SUCCESS");
      write(sockfd, data, sizeof(data));
    }

  } else if (CLOSE_DOWNSIDE_FIFO == recvCommand) {

#ifdef DEBUG
    cerr << FIFOManagerSIG << "connectSelector: CLOSE_DOWNSIDE_FIFO comman" << endl;
#endif
    // ID
    begIdx = recvData.find_first_not_of(delims, endIdx);
    if (begIdx == std::string::npos) {
      cerr << "not recv id" << endl;
      //return -1;
      pthread_exit(NULL);
    }
    endIdx = recvData.find_first_of(delims, begIdx);
    if (endIdx == std::string::npos) {
      endIdx = recvData.length();
    }
    std::string id = recvData.substr(begIdx, (endIdx-begIdx));
    cerr << "id=[" << id << "]" << endl;

    int result = that->closeDownsideFIFO(id);
    if (0 != result) {
      sprintf(data, "FAILURE 104");
      write(sockfd, data, sizeof(data));
    } else {
      sprintf(data, "SUCCESS");
      write(sockfd, data, sizeof(data));
    }

  } else {
    cerr << "invalid command" << endl;
  }

  //sleep(10);
  cerr << "bye.." <<endl;

  close(sockfd);
  pthread_exit(NULL);
}

void FIFOManager::makeFIFOInfo(std::list<string> fifoColl)
{
  //(P)擪̗vfo
  //(Q)̗vfƑ΂ɂȂvf
  //(R)΂̗vfȂ,̗vfjĎ̗vf擾(P)
  //(S)΂̗vf΂̗RNV폜
  //(T)΂̗vfeFIFOInfoCX^X𐶐,RNVɒǉ

  std::string fifoYou;  //΂ƂȂvf(FIFOt@C)
  std::string fifoMe;   //Jgvf(FIFOt@C)
  std::string fifoPath; //FIFOt@C̃pX(fBNg)
  std::string commonId; //΂̗vfɋʂID

  std::list<string>::iterator pos;
  //for (pos = fifoColl.begin(); pos != fifoColl.end(); ++pos) {}
  pos = fifoColl.begin();
  while (pos != fifoColl.end()) {
    fifoMe = *pos;  // ł̓t@CtpX
    fifoPath = *pos;
    //pX̕ۑ
    std::string::size_type idx = fifoPath.rfind(FILE_PATH_DELIMIT);
    //size_t idx = fifoPath.rfind('/');
    fifoPath.erase(idx+1);
    //pX̏
    //size_type idx = fifoMe.rfind('/');
    fifoMe.erase(0, idx+1);
    commonId = fifoMe;
    pos = fifoColl.erase(pos);

#ifdef DEBUG
    cerr << "*** fifoMe=[" << fifoMe << "] commonId=[" << commonId << "] ***" << endl;
#endif

    std::string comString(fifoMe.substr(0, FIFOInfo::UPSIDE_FIFO_PREFIX.length()));
    if (!comString.compare(FIFOInfo::UPSIDE_FIFO_PREFIX)) {
      // DOWNSIDE_FIFO_PREFIX
      commonId.erase(0, FIFOInfo::UPSIDE_FIFO_PREFIX.length());
#ifdef DEBUG
      cerr << "## 1 ## ID=[" << commonId << "] YOU=[" << fifoYou << "]" << endl;
#endif
      fifoYou = fifoPath + FIFOInfo::DOWNSIDE_FIFO_PREFIX + commonId;
      //fifoYou = fifoMe.replace(0, UPSIDE_FIFO_PREFIX.length(), DOWNSIDE_FIFO_PREFIX);
    } else {
      // UPSIDE_FIFO_PREFIX
      commonId.erase(0, FIFOInfo::DOWNSIDE_FIFO_PREFIX.length());
#ifdef DEBUG
      cerr << "## 2 ## ID=[" << commonId << "] YOU=[" << fifoYou << "]" << endl;
#endif
      fifoYou = fifoPath + FIFOInfo::UPSIDE_FIFO_PREFIX + commonId;
      //fifoYou = fifoMe.replace(0, DOWNSIDE_FIFO_PREFIX.length(), UPSIDE_FIFO_PREFIX);
    }
#ifdef DEBUG
    cerr << FIFOManagerSIG << "makeFIFOInfo: You:[" << fifoYou
	 << "] and Me:[" << fifoMe << "]" << endl;
#endif

    std::list<string>::iterator pairPos = find(pos, fifoColl.end(), fifoYou);
    if (pairPos != fifoColl.end()) {
#ifdef DEBUG
      cerr << "*** FIND ***" << endl;
#endif
      //΂ƂȂvf
      //fifoInfoColl_.push_back(FIFOInfo(commonId));
      FIFOInfo fifoInfo(commonId, fifoPath+fifoMe, fifoYou);
      //͑dXbhɓOȂ̂mutex_lock̕Kv
      fifoInfoColl_.push_back(fifoInfo);
    }
  }

#ifdef DEBUG
  {
    cerr << "FIFOManager:: make FIFOInfo list" << endl;
    cerr << "-----------------------------" << endl;
    std::list<FIFOInfo>::const_iterator pos;
    for (pos = fifoInfoColl_.begin(); pos != fifoInfoColl_.end(); ++pos) {
      pos->spy();
    }
    cerr << "-----------------------------" << endl;
  }
#endif

}

//------------------------------------------------------------
// fXgN^
//------------------------------------------------------------
FIFOManager::~FIFOManager()
{
  // ϐj
  int status = pthread_cond_destroy(&fifoInfoColl_cond_);
  if (0 != status) {
    //throw OpenSOAPException();
    //throw std::exception();
    cerr << FIFOManagerSIG << "~FIFOManager: mutext_destroy failed" << endl;
  }
  // mutex j
  status = pthread_mutex_destroy(&fifoInfoColl_mutex_);
  if (0 != status) {
    //throw OpenSOAPException();
    //throw std::exception();
    cerr << FIFOManagerSIG << "~FIFOManager: mutext_destroy failed" << endl;
  }
}

int FIFOManager::closeUpsideFIFO(const std::string& id)
{
  std::list<FIFOInfo>::iterator pos;
  // mutex lock
  // bNݒ
  int status = pthread_mutex_lock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return -1;
  }
  
  for (pos = fifoInfoColl_.begin(); pos != fifoInfoColl_.end(); ++pos) {
    if (pos->getId() == id) {
      break;
    }
  }

  int retStatus = 0;
  if (pos != fifoInfoColl_.end()) {
    // IDɑΉ̂Ȃ΍Ō܂ŃV[NĂ
    if (pos->getStatus() == FIFOInfo::DOWNSIDE_CLOSED 
	|| pos->getStatus() == FIFOInfo::UPSIDE_READY) {
      pos->setStatus(FIFOInfo::INIT);
      //󂫑҂ɍ}𑗂
      int status = pthread_cond_signal(&fifoInfoColl_cond_);
      if (0 != status) {
	cerr << TAG_ERR << FIFOManagerSIG << "closeUpsideFIFO: cond signal failed" << endl;
      }

    } else if (pos->getStatus() == FIFOInfo::DOWNSIDE_CONNECTING) {
      // warning !!
      // DOWNSIDE(NCAg)ڑ
      //pos->setStatus(FIFOInfo::INIT);
      pos->setStatus(FIFOInfo::UPSIDE_CLOSED);
      //retStatus = 1;

    } else {
      cerr << TAG_ERR << FIFOManagerSIG << "closeUpsideFIFO: invalid FIFOInfo status:[" 
	   << pos->getStatus() << "]" << endl;
      retStatus = -1;
    }
  }

  // bN
  status = pthread_mutex_unlock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return -1;
  }

  return retStatus;
}

int FIFOManager::closeDownsideFIFO(const std::string& id)
{
  std::list<FIFOInfo>::iterator pos;
  // mutex lock
  // bNݒ
  int status = pthread_mutex_lock(&fifoInfoColl_mutex_);
  if (0 != status) {
    cerr << FIFOManagerSIG << "closeDownsideFIFO: mutex lock failed" << endl;
    //error
    return -1;
  }

  for (pos = fifoInfoColl_.begin(); pos != fifoInfoColl_.end(); ++pos) {
    if (pos->getId() == id) {
      break;
    }
  }

  int retStatus = 0;
  if (pos != fifoInfoColl_.end()) {
    // IDɑΉ̂Ȃ΍Ō܂ŃV[NĂ
    if (pos->getStatus() == FIFOInfo::DOWNSIDE_CONNECTING) {
#ifdef DEBUG
      cerr << "*** 1 ***" << endl;
#endif
      //pos->setStatus(FIFOInfo::UPSIDE_READY);
      pos->setStatus(FIFOInfo::DOWNSIDE_CLOSED);
      
    } else if (pos->getStatus() == FIFOInfo::UPSIDE_CLOSED) {
#ifdef DEBUG
      cerr << "*** 2 ***" << endl;
#endif
      pos->setStatus(FIFOInfo::INIT);
      //󂫑҂ɍ}𑗂
      int status = pthread_cond_signal(&fifoInfoColl_cond_);
      if (0 != status) {
	cerr << TAG_ERR << FIFOManagerSIG << "closeUpsideFIFO: cond signal failed" << endl;
      }

    } else if (pos->getStatus() == FIFOInfo::INIT) {
      cerr << TAG_ERR << FIFOManagerSIG 
	   << "closeDownsideFIFO: invalid FIFOInfo status:[INIT]" << endl;

#ifdef DEBUG
      cerr << "*** 3 ***" << endl;
#endif
      //upsideclose

    } else {
      cerr << TAG_ERR << FIFOManagerSIG << "closeDownsideFIFO: invalid FIFOInfo status:["
	   << pos->getStatus() << "]" << endl;
#ifdef DEBUG
      cerr << "*** 4 ***" << endl;
#endif
      //retStatus = -1;
    }
  }

  // bN
  status = pthread_mutex_unlock(&fifoInfoColl_mutex_);
  if (0 != status) {
    cerr << FIFOManagerSIG << "closeDownsideFIFO: mutex unlock failed" << endl;
    //error
    return -1;
  }

  return retStatus;

}

FIFOManager::FIFOInfo* FIFOManager::openUpsideFIFO(const int timeout_second)
{
  std::list<FIFOInfo>::iterator pos;
  // mutex lock
  // bNݒ
  int status = pthread_mutex_lock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return NULL;
  }
  FIFOInfo* fifoInfo = NULL;
  while (true) {
    fifoInfo = NULL;
    for (pos = fifoInfoColl_.begin(); pos != fifoInfoColl_.end(); ++pos) {
      if (pos->getStatus() == FIFOInfo::INIT) {
	pos->setStatus(FIFOInfo::UPSIDE_READY);
	fifoInfo = &(*pos);

	//󂫑҂ɍ}𑗂
	int status = pthread_cond_signal(&fifoInfoColl_cond_);
	if (0 != status) {
	  cerr << TAG_ERR << FIFOManagerSIG << "openUpsideFIFO: cond signal failed" << endl;
	}

	break; // exit for ();
      }
    }
    if (NULL != fifoInfo) {
      //󂫂!!
      break; //exit while (true);
    }

    //󂭂܂ő҂
    //    if (NO_WAIT != timeout_second) {
      struct timespec timeout;
      int timeoutSecond = 0;
      if (WAIT_NOLIMIT == timeout_second) {
	timeoutSecond = 180;
      } else if (timeout_second > 0) {
	timeoutSecond = timeout_second;
      } else if (NO_WAIT == timeout_second) {
	timeoutSecond = 2;
      } else {
	cerr << TAG_ERR << FIFOManagerSIG << "openUpsideFIFO: invalid timeout second" << endl;
      }
      
      timeout.tv_sec = time(NULL) + timeoutSecond;
      timeout.tv_nsec = 0;

#ifdef DEBUG
      cerr << "FIFOManager::openUpsideFIFO: cond_timedwait(" << timeoutSecond << ")" << endl;
#endif

      //Ԃς邩,w莞ԃ^CAEg܂ő҂
      int status = pthread_cond_timedwait(&fifoInfoColl_cond_, 
					   &fifoInfoColl_mutex_,
					   &timeout);

      if (ETIMEDOUT == status) {
#ifdef DEBUG
	cerr << FIFOManagerSIG << "openUpsideFIFO: open wait timeouted:[" 
	     << timeoutSecond << "sec.]" << endl;
#endif
	break;
	
      } else if (0 != status) {
	cerr << TAG_ERR << FIFOManagerSIG << "openUpsideFIFO: wait cond timeout failed" << endl;
      }
      //    }
  }
  // mutex unlock
  // bN
  status = pthread_mutex_unlock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return NULL;
  }

#if 0
  if (pos != fifoInfoColl_.end()) {
    // INITXe[^X̂̂Ȃ΍Ō܂ŃV[NĂ
    //int r_fd = open(pos->getUpsideFIFO().c_str(), O_RDONLY);
    //int w_fd = open(pos->getDownsideFIFO().c_str(), O_WRONLY);

    //pos->getUpsideFIFODescriptor().readFIFO_ = r_fd;
    //pos->getUpsideFIFODescriptor().writeFIFO_ = w_fd;
    //pos->setStatus(FIFOInfo::BUSY);
    
    //fifo_desc = pos->getUpsideFIFODescriptor();
    fifoInfo = &(*pos);
  }
#endif

  return fifoInfo;
}

FIFOManager::FIFOInfo* FIFOManager::openDownsideFIFO(const int timeout_second)
{
  std::list<FIFOInfo>::iterator pos;
  // mutex lock
  // bNݒ
  int status = pthread_mutex_lock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return NULL;
  }
  FIFOInfo* fifoInfo = NULL;
  while (true) {
    fifoInfo = NULL;
    for (pos = fifoInfoColl_.begin(); pos != fifoInfoColl_.end(); ++pos) {
      if (pos->getStatus() == FIFOInfo::UPSIDE_READY) {
	pos->setStatus(FIFOInfo::DOWNSIDE_CONNECTING);
	fifoInfo = &(*pos);
	break;
      }
    }
    if (NULL != fifoInfo) {
      //󂫂!!
      break; //exit while (true);
    }

    //󂭂܂ő҂
    if (NO_WAIT != timeout_second) {
      struct timespec timeout;
      int timeoutSecond = 0;
      if (WAIT_NOLIMIT == timeout_second) {
	timeoutSecond = 180;
      } else if (timeout_second > 0) {
	timeoutSecond = timeout_second;
	//      } else if (NO_WAIT == timeout_second) {
	//	timeoutSecond = 2;
      } else {
	cerr << TAG_ERR << FIFOManagerSIG << "openDownsideFIFO: invalid timeout second" << endl;
      }
      
      timeout.tv_sec = time(NULL) + timeoutSecond;
      timeout.tv_nsec = 0;

#ifdef DEBUG
      cerr << "FIFOManager::openDownsideFIFO: cond_timedwait(" << timeoutSecond << ")" << endl;
#endif

      //Ԃς邩,w莞ԃ^CAEg܂ő҂
      int status = pthread_cond_timedwait(&fifoInfoColl_cond_, 
					   &fifoInfoColl_mutex_,
					   &timeout);

      if (ETIMEDOUT == status) {
#ifdef DEBUG
	cerr << FIFOManagerSIG << "openDownsideFIFO: open wait timeouted:[" 
	     << timeoutSecond << "sec.]" << endl;
#endif
	break;
	
      } else if (0 != status) {
	cerr << TAG_ERR << FIFOManagerSIG << "openDownsideFIFO: wait cond timeout failed" << endl;
      }
    } else {
      break;
    }
  }

  // mutex unlock
  // bN
  status = pthread_mutex_unlock(&fifoInfoColl_mutex_);
  if (0 != status) {
    //error
    return NULL;
  }

#if 0
  if (pos != fifoInfoColl_.end()) {
    // INITXe[^X̂̂Ȃ΍Ō܂ŃV[NĂ
    //int w_fd = open(pos->getUpsideFIFO().c_str(), O_WRONLY);
    //int r_fd = open(pos->getDownsideFIFO().c_str(), O_RDONLY);

    //pos->getDownsideFIFODescriptor().readFIFO_ = r_fd;
    //pos->getDownsideFIFODescriptor().writeFIFO_ = w_fd;
    //pos->setStatus(FIFOInfo::BUSY);

    //fifo_desc = pos->getDownsideFIFODescriptor();

    fifoInfo = &(*pos);
  }
#endif

  return fifoInfo;
}


//========== FIFOInfo =============
//FIFOZbg(艺)Ǘ邽߂̃f[^NX

/*
      enum FIFO_STAT { INIT = 0,      //
		       UPSIDE_READY,  //󂯑ڑ
		       UPSIDE_CLOSED, //󂯑Ǐ
		       BUSY };        //ڑ
*/

const std::string FIFOManager::FIFOInfo::UPSIDE_FIFO_PREFIX = "_UPSIDE_";
const std::string FIFOManager::FIFOInfo::DOWNSIDE_FIFO_PREFIX = "_DOWNSIDE_";


void FIFOManager::FIFOInfo::spy() const
{
  cerr << "FIFOInfo::spy: STAT=[" << fifoStat_ << "]: ID=[" << upsideDesc_.getId() << "]" << endl;
}

//FIFÕt@CKɏ]Ă邩̃`FbN
bool FIFOManager::FIFOInfo::checkValidFileName(const std::string& fifoFile)
{
  std::string compString(fifoFile.substr(0, UPSIDE_FIFO_PREFIX.length()));
  return (!compString.compare(UPSIDE_FIFO_PREFIX) ||
	  !compString.compare(DOWNSIDE_FIFO_PREFIX))
    ? true : false;
}

FIFOManager::FIFOInfo::FIFOInfo(const std::string& id, 
				const std::string& fifoUpside,
				const std::string& fifoDownside)
  : fifoStat_(INIT)
  , upsideDesc_(id)
  , downsideDesc_(id)
  , fifoUpside_(fifoUpside)
  , fifoDownside_(fifoDownside)
{
  
}

FIFOManager::FIFOInfo::~FIFOInfo()
{
  
}

// End of FIFOManager.cpp
