/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: FwdInfoFunc.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include "ServerCommon.h"
#else
#include <unistd.h>
#include <netdb.h>
#endif

#include <algorithm>
#include <iostream>
#include <string>

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif /* HAVE_CONFIG_H */

/* libxml2 */
#if defined(HAVE_LIBXML2_XMLVERSION_H) 
#include <libxml2/tree.h>
#include <libxml2/parser.h>
#else
#include <libxml/tree.h>
#include <libxml/parser.h>
#endif /* !HAVE_LIBXML2_XMLVERSION_H */

#include "FwdInfoFunc.h"

#include "SrvConfAttrHandler.h"
#include "SOAPMessage.h"

using namespace OpenSOAP;
using namespace std;

//----- Forward ftHg̃zXg̏擾 -----//
bool
getForwarderInfo(std::string& forwardHostname, unsigned int& forwardPort, 
		 std::string& forwardCGILocation)
{
  SrvConfAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  std::string query;
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/forwarder/hostname=?";
#else
  query = SERVER_CONF_TAG + "/forwarder/hostname=?";
#endif
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }

#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
#endif

  forwardHostname = values[0];
  
  values.clear();
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/forwarder/port=?";
#else
  query = SERVER_CONF_TAG + "/forwarder/port=?";
#endif
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
#endif
  forwardPort = atoi(values[0].c_str());
  
  values.clear();
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/forwarder/cgi_location=?";
#else
  query = SERVER_CONF_TAG + "/forwarder/cgi_location=?";
#endif
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getForwarderInfo:" 
       << "values[0] = " << values[0] << endl;
#endif
  forwardCGILocation = values[0];
  
  std::string officialHostname = getLocalhostName();
  
  if (officialHostname == forwardHostname) {
#ifdef DEBUG
    cerr << "Official hostname equals to forward hostname." << endl;
    cerr << "This will cause loop problem !!" << endl;
#endif
    return false;
  }
  
  return true;
}// end of getForwarderInfo(std::string&, unsigned int&, std::string&)


//----- Forward ꂽbZ[WԂĂ炤 URL 擾 -----//
bool
getBackwardURLs(std::vector<std::string>& backwardURLs)
{
  SrvConfAttrHandler ssmlAttrHndl;
  std::vector<std::string> values;
  std::string query;

  std::string protocol;
  std::string port;
  std::string cgiLocation;
  
  // vgR̎擾
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/backward/protocol=?";
#else
  query = SERVER_CONF_TAG + "/backward/protocol=?";
#endif
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" 
       << "values[0] = " << values[0] << endl;
#endif

  protocol = values[0];
  
  // |[gԍ̎擾
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/backward/port=?";
#else
  query = SERVER_CONF_TAG + "/backward/port=?";
#endif
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
#ifdef DEBUG
    cerr << "*** ERROR ***" << endl;
    cerr << "Use default port: 80" << endl;
#endif

    port = "80";
  }
  else {
#ifdef DEBUG
    cerr << "### RESULT ###" << endl;
    cerr << "FwdInfoFunc::getBackwardURLs:" 
	 << "values[0] = " << values[0] << endl;
#endif
    port = values[0];
  }
  
  
  // CGI ւ̃pX̎擾
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/backward/cgi_location=?";
#else
  query = SERVER_CONF_TAG + "/backward/cgi_location=?";
#endif
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" 
       << "values[0] = " << values[0] << endl;
#endif
  cgiLocation = values[0];


  // get hostname
#if !defined(WIN32)
  query = SrvConfAttrHandler::SERVER_CONF_TAG + "/backward/hostname=??";
#else
  query = SERVER_CONF_TAG + "/backward/hostname=??";
#endif
  values.clear();
  if (0 > ssmlAttrHndl.queryXml(query, values)) {
    cerr << "*** ERROR ***" << endl;
    return false;
  }
  
#ifdef DEBUG
  cerr << "### RESULT ###" << endl;
  cerr << "FwdInfoFunc::getBackwardURLs:" << endl;
#endif

  std::vector<std::string>::iterator iter;

#ifdef DEBUG
  for(iter = values.begin(); iter != values.end(); iter++) {
    cerr << "values = " << *iter << endl;
  }
#endif

  // make backwardURLs
  if (port == "80") {
    for(iter = values.begin(); iter != values.end(); iter++) {    
      backwardURLs.push_back(protocol + "://" 
			     + *iter + cgiLocation);
    }
  }
  else {
    for(iter = values.begin(); iter != values.end(); iter++) {
      backwardURLs.push_back(protocol + "://" 
			    + *iter + ":" + port + cgiLocation);
    }
  }
#ifdef DEBUG
  //for (std::vector<std::string>::iterator iter = backwardURLs.begin();
  for (iter = backwardURLs.begin(); iter != backwardURLs.end(); iter++) {
    cerr << "backwarkURLs = [" << *iter << "]" <<endl;
  }
#endif

  return true;
}// end of getBackwardURLs(std::vector<std::string>&)



//-- SOAP Header ɂ forward_path ̓e؂蕪 --//
bool
analyseFwdHostInfo(std::vector<std::string> fwdPath,
		   std::vector<std::string>& prtcls,
		   std::vector<std::string>& hostnames,
		   std::vector<unsigned int>& ports,
		   std::vector<std::string>& cgiLocations)
{
  prtcls.clear();
  hostnames.clear();
  ports.clear();
  cgiLocations.clear();
  
  std::string::iterator findPoint;
  std::string* tmp;
  std::string* lastStr;
  char delim;
  
  for (std::vector<std::string>::iterator iter = fwdPath.begin();
      iter != fwdPath.end(); iter++) {
    /* vgR擾 */
    delim = ':';
    findPoint = find(iter->begin(), iter->end(), delim);
    if (findPoint != iter->end()) {
#if 1 //2003/02/06
      prtcls.push_back(std::string(iter->begin(), findPoint));
#else
      tmp = new std::string(iter->begin(), findPoint);
#ifdef DEBUG
      cout << "*tmp = " << *tmp << endl;
#endif

      prtcls.push_back(*tmp);
#endif //if 1

      if (findPoint < iter->end() - 1) {
		lastStr = new std::string(findPoint + 1, iter->end());
#ifdef DEBUG
		cout << "*lastStr = " << *lastStr << endl;
#endif
      }
      else {
	return false;
      }
    }
    else {
      return false;
    }
    
    /* zXgƃ|[gԍ擾 */
    // '/'  2 
    for (unsigned int i = 0; i < 2; i++) {
      delim = '/';
      findPoint = find(lastStr->begin(), lastStr->end(), delim);
      if (findPoint != lastStr->end()) {
	if (findPoint < lastStr->end() - 1) {
	  lastStr = new std::string(findPoint + 1, lastStr->end());
#ifdef DEBUG
	  cout << "*lastStr = " << *lastStr << endl;
#endif
	}
	else {
	  return false;
	}
      }
      else {
	return false;
      }
    }// loop of for (unsigned int i = 0; i < 2; i++) 
    
    
    bool useDefaultPort = true;
    delim = ':';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      tmp = new std::string(lastStr->begin(), findPoint);
#ifdef DEBUG
      cout << "*tmp = " << *tmp << endl;
#endif
      useDefaultPort = false;
      hostnames.push_back(*tmp);
      if (findPoint < lastStr->end() - 1) {
		lastStr = new std::string(findPoint + 1, lastStr->end());
#ifdef DEBUG
		cout << "*lastStr = " << *lastStr << endl;
#endif
      }
      else {
	return false;
      }
    }
    else {
#ifdef DEBUG
      cout << "using default port: 80" << endl;
#endif
      ports.push_back(80);
    }
    
    delim = '/';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      tmp = new std::string(lastStr->begin(), findPoint);
#ifdef DEBUG
      cout << "*tmp = " << *tmp << endl;
#endif

      if (useDefaultPort) {
	hostnames.push_back(*tmp);
      }
      else {
	ports.push_back((unsigned int)atoi(tmp->c_str()));
      }
      if (findPoint < lastStr->end() - 1) {
		/* CGI ւ̃pX擾 */
		lastStr = new std::string(findPoint, lastStr->end());
		cgiLocations.push_back(*lastStr);
#ifdef DEBUG
		cout << "cgiLocation = " << *lastStr << endl;
#endif
      }
      else {
	return false;
      }
    }
    else {
      return false;
    }
    
    
  }// loop of (iter)
#ifdef DEBUG
  cout << "----- fwdPathInfo -----" << endl;
  for (unsigned int i = 0; i < fwdPath.size(); i++) {
    cout << "prtcls[" << i << "] = " << prtcls[i] << endl;
    cout << "hostnames[" << i << "] = " << hostnames[i] << endl;
    cout << "ports[" << i << "] = " << ports[i] << endl;
    cout << "cgiLocations[" << i << "] = " << cgiLocations[i] << endl;
    cout << endl;
  }
  cout << "----- fwdPathInfo -----" << endl;
#endif

  return true;
}// end of analyseFwdHostInfo()

//-- SOAP Header ɂ forward_path ̓e؂蕪 --//
extern
bool
analyseFwdHostInfo(std::string fwdPath,
		   std::string& prtcl, std::string& hostname,
		   unsigned int& port, std::string& cgiLocation)
{
  
  prtcl = "";
  hostname = "";
  port = 80;
  cgiLocation = "";
  
  std::string::iterator findPoint;
  std::string* tmp;
  std::string* lastStr;
  char delim;
  
  /* vgR擾 */
  delim = ':';
  findPoint = find(fwdPath.begin(), fwdPath.end(), delim);
  if (findPoint != fwdPath.end()) {
    tmp = new std::string(fwdPath.begin(), findPoint);
#ifdef DEBUG
    cout << "*tmp = " << *tmp << endl;
#endif
    prtcl = *tmp;
    if (findPoint < fwdPath.end() - 1) {
      lastStr = new std::string(findPoint + 1, fwdPath.end());
#ifdef DEBUG
      cout << "*lastStr = " << *lastStr << endl;
#endif
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
  
  /* zXgƃ|[gԍ擾 */
  // '/'  2 
  for (unsigned int i = 0; i < 2; i++) {
    delim = '/';
    findPoint = find(lastStr->begin(), lastStr->end(), delim);
    if (findPoint != lastStr->end()) {
      if (findPoint < lastStr->end() - 1) {
		lastStr = new std::string(findPoint + 1, lastStr->end());
#ifdef DEBUG
		cout << "*lastStr = " << *lastStr << endl;
#endif
      }
	else {
	  return false;
	}
    }
      else {
	return false;
      }
  }// loop of for (unsigned int i = 0; i < 2; i++) 
  
  
  bool useDefaultPort = true;
  delim = ':';
  findPoint = find(lastStr->begin(), lastStr->end(), delim);
  if (findPoint != lastStr->end()) {
    tmp = new std::string(lastStr->begin(), findPoint);
#ifdef DEBUG
    cout << "*tmp = " << *tmp << endl;
#endif
    useDefaultPort = false;
    hostname = *tmp;
    if (findPoint < lastStr->end() - 1) {
      lastStr = new std::string(findPoint + 1, lastStr->end());
#ifdef DEBUG
      cout << "*lastStr = " << *lastStr << endl;
#endif
    }
    else {
      return false;
    }
  }
  else {
#ifdef DEBUG
    cout << "using default port: 80" << endl;
#endif
    port = 80;
  }
  
  delim = '/';
  findPoint = find(lastStr->begin(), lastStr->end(), delim);
  if (findPoint != lastStr->end()) {
    tmp = new std::string(lastStr->begin(), findPoint);
#ifdef DEBUG
    cout << "*tmp = " << *tmp << endl;
#endif

    if (useDefaultPort) {
      hostname = *tmp;
    }
    else {
      port = (unsigned int)atoi(tmp->c_str());
    }
    if (findPoint < lastStr->end() - 1) {
      /* CGI ւ̃pX擾 */
      lastStr = new std::string(findPoint, lastStr->end());
      cgiLocation = *lastStr;
#ifdef DEBUG
      cout << "cgiLocation = " << *lastStr << endl;
#endif
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
  
  
  
#ifdef DEBUG
  cout << "----- fwdPathInfo -----" << endl;
  cout << "prtcl = " << prtcl << endl;
  cout << "hostname = " << hostname << endl;
  cout << "port = " << port << endl;
  cout << "cgiLocation = " << cgiLocation << endl;
  cout << endl;
  
  cout << "----- fwdPathInfo -----" << endl;
#endif

  return true;
}

