/* -*- mode: c++; -*-
 *-----------------------------------------------------------------------------
 * $RCSfile: SrvDrvCreator_main.cpp,v $
 *
 * See Copyright for the status of this software.
 *
 * The OpenSOAP Project
 * http://opensoap.jp/
 *-----------------------------------------------------------------------------
 */

#if defined(WIN32)
#include <windows.h>
#include <io.h>
#else
#include <unistd.h>
#endif

#include <iostream>
#include <fstream>
#include <string>

#if !defined(NO_LOGGER)
#include "Logger.h"
#endif
#include "ServerCommon.h"
#include "SrvDrvCreator.h"
#include "PIDFactory.h"
#include "SrvConfAttrHandler.h"

using namespace OpenSOAP;
using namespace std;

int 
main(int argc, char** argv) {
  bool isAsync = false;
  
  // It works as Asynchronous Service Driver, 
  // except for argument is 0 or no argument exists

  // initial argument differ in windows and other.
#if defined(WIN32)
  // windows argument : program logfile [isAsyncFlg]
  //string logFile = OPENSOAP_TMP_PATH + "log/srvDrvCreator_sync";
  if (argc > 2 && (atoi(argv[2]) != 0)) {
    isAsync = true;
  }
#else
  // other argument : program [isAsyncFlg]
  if (argc > 1 && (atoi(argv[1]) != 0)) {
	  isAsync = true;
  }
#endif

#if defined(NO_LOGGER)
  string logFile;
  SrvConfAttrHandler* srvConf = new SrvConfAttrHandler();
  vector<string> values;
  string query = SERVER_CONF_TAG + "/log/path=?";
  if (0 > srvConf->queryXml(query, values)) {
    if (isAsync) {
      logFile = OPENSOAP_TMP_PATH + "log/srvDrvCreator_async";
    }
    else {
      logFile = OPENSOAP_TMP_PATH + "log/srvDrvCreator_sync";
    }
  }
  else {
    if (isAsync) {
      logFile = values[0] + "srvDrvCreator_async.log";
    }
    else {
      logFile = values[0] + "srvDrvCreator_sync.log";
    }
  }
  delete srvConf;

  FILE* outfp = fopen(logFile.c_str(), "w");
  if (!outfp) {
    return 1;
  }
  if (argc <=1) {
    fprintf(outfp, "invalid initial argument. argc=(%d)\n", argc);
    fprintf(outfp, "exit.\n");
    fclose(outfp);
    return EXIT_FAILURE;
  }

  int old;
  int olderr;
  //backup stdout
  old = _dup(1);
  //stdout -> outfp
  if (-1 == _dup2(_fileno(outfp), 1)) {
    fprintf(outfp, "_dup2 to stdout failed.\n");
    fprintf(outfp, "exit.\n");
    fclose(outfp);
    return EXIT_FAILURE;
  }
  //backup stderr
  olderr = _dup(2);
  //stderr -> outfp
  if (-1 == _dup2(_fileno(outfp), 2)) {
    fprintf(outfp, "_dup2 to stderr failed.\n");
    fprintf(outfp, "exit.\n");
    fclose(outfp);
    return EXIT_FAILURE;
  }
#else
  // set log level  
  Logger::InitOpt opt;
  opt.outputLevel = LOG_DEBUG;
  Logger::init(opt);
  //write start message
  if (isAsync) {
    Logger::printLog(NULL, " SrvDrvCreator_main (async) start", LOG_INFO,
		     0x00100003L, "exe");
  }
  else {
    Logger::printLog(NULL, " SrvDrvCreator_main (sync) start", LOG_INFO,
		     0x00100003L, "exe");
  }
#endif

#if !defined(WIN32)
  // make it to daemon & write pid file

  daemon(0, 1);


  PIDFactory* pidFactory = new PIDFactory();
  pidFactory->makePID(isAsync ? ASYNC_SRVDRV_PID_FILE : SYNC_SRVDRV_PID_FILE);
  delete pidFactory;

#endif

  // create instance of QueueManager and run
  ChannelManager* chnlMngr = new SrvDrvCreator(isAsync);
  chnlMngr->run();

  delete chnlMngr;

#if defined(NO_LOGGER)
  fclose(outfp);
  _dup2(old, 1);
  _dup2(olderr, 2);
#else
  //write stop message
  if (isAsync) {
    Logger::printLog(NULL, " SrvDrvCreator_main (async) stop", LOG_INFO,
		     0x00100003L, "exe");
  }
  else {
    Logger::printLog(NULL, " SrvDrvCreator_main (sync) stop", LOG_INFO,
		     0x00100003L, "exe");
  }
  Logger::term();
#endif

  return EXIT_SUCCESS;
}
